#!/usr/bin/env python3

'''Unix domain socket connection and manager.'''

import asyncio

from aiospamc.connections import Connection
from aiospamc.connections import ConnectionManager
from aiospamc.exceptions import AIOSpamcConnectionFailed


class UnixConnectionManager(ConnectionManager):
    '''Creates new connections based on Unix domain socket path provided.

    Attributes
    ----------
    path : str
        Path of the socket.
    '''

    def __init__(self, path, loop=None):
        '''Constructor for UnixConnectionManager.

        Parameters
        ----------
        path : str
            Path of the socket.
        loop : asyncio.AbstractEventLoop
            The asyncio event loop.
        '''

        self.path = path
        super().__init__(loop)

    def __repr__(self):
        return 'UnixConnectionManager(path={})'.format(repr(self.path))

    def new_connection(self):
        '''Creates a new Unix domain socket connection.

        Raises
        ------
        AIOSpamcConnectionFailed
        '''

        return UnixConnection(self.path, self.loop)


class UnixConnection(Connection):
    '''Manages a Unix domain socket connection.

    Attributes
    ----------
    path : str
        Path of the socket.
    loop : asyncio.AbstractEventLoop
        The asyncio event loop.
    '''

    def __init__(self, path, loop=None):
        '''Constructor for UnixConnection.

        Parameters
        ----------
        path : str
            Path of the socket.
        loop : asyncio.AbstractEventLoop
            The asyncio event loop.
        '''

        self.path = path
        super().__init__(loop)

    def __repr__(self):
        return 'UnixConnection(path={})'.format(repr(self.path))

    async def open(self):
        '''Opens a connection.

        Returns
        -------
        asyncio.StreamReader
        asyncio.StreamWriter

        Raises
        ------
        aiospamc.exceptions.AIOSpamcConnectionFailed
        '''

        try:
            reader, writer = await asyncio.open_unix_connection(self.path,
                                                                loop=self.loop)
        except (ConnectionRefusedError, OSError) as error:
            raised = AIOSpamcConnectionFailed(error)
            self.logger.exception('Exception occurred when connecting: %s', raised)
            raise raised

        return reader, writer

    @property
    def connection_string(self):
        '''String representation of the connection.

        Returns
        -------
        str
            Path to the Unix domain socket.
        '''

        return self.path



#!/usr/bin/env python3

'''Common classes for the project.'''

import zlib

from aiospamc.headers import ContentLength


class RequestResponseBase:
    '''Base class for requests and responses.'''

    def __init__(self, body=None, headers=None):
        '''
        Parameters
        ----------
        body : :obj:`str`, optional
            String representation of the body.  An instance of the
            aiospamc.headers.ContentLength will be automatically added.
        headers : tuple of :class:`aiospamc.headers.Header`, optional
            Collection of headers to be added.  If it contains an instance of
            aiospamc.headers.Compress then the body is automatically
            compressed.
        '''

        if headers:
            self._headers = {item.field_name(): item for item in headers}
        else:
            self._headers = {}
        self._body = ''
        self._compressed_body = None
        if body:
            if isinstance(body, str):
                self.body = body
            elif isinstance(body, bytes):
                self.body = self._decode_body(body, self._headers.values())

    @staticmethod
    def _decode_body(body, headers):
        '''Parses a body of a response.

        Parameters
        ----------
        body : bytes
            Bytes representation of body.
        headers : :obj:`tuple` or :obj:`list` of :class:`aiospamc.headers.Header`
            Collection of headers.

        Returns
        -------
        str
            Body in a string object.
        '''

        if not body:
            return None
        elif any(header.field_name() == 'Compress' for header in headers):
            return zlib.decompress(body).decode()
        else:
            return body.decode()

    @property
    def body(self):
        '''Contains the contents of the body.

        The getter will return a bytes object.

        The setter expects a string.  If the :class:`aiospamc.headers.Compress`
        header is present then the value of body will be compressed.

        The deleter will automatically remove the
        :class:`aiospamc.headers.ContentLength` header.
        '''

        return self._body

    @body.setter
    def body(self, value):
        self._body = value
        if 'Compress' in self._headers:
            self._compress_body()
        self._set_content_length(value)

    @body.deleter
    def body(self):
        self._body = ''
        self._compressed_body = None
        self.delete_header('Content-length')

    def _compress_body(self):
        self._compressed_body = zlib.compress(self.body.encode())
        self._set_content_length(self._compressed_body)

    def _decompress_body(self):
        self._compressed_body = None
        self._set_content_length(self.body)

    def _set_content_length(self, body_):
        self.add_header(ContentLength(len(body_)))

    def add_header(self, header):
        '''Adds a header to the request.  A header with the same name will be
        overwritten.

        Parameters
        ----------
        header : :class:`aiospamc.headers.Header`
            A header object to be added.
        '''

        if header.field_name() == 'Compress' and self.body:
            self._compress_body()
        self._headers[header.field_name()] = header

    def get_header(self, header_name):
        '''Gets the header matching the name.

        Parameters
        ----------
        header_name : :obj:`str`
            String name of the header.

        Returns
        -------
        :class:`aiospamc.headers.Header`
            A Header object or subclass of it.

        Raises
        ------
        :class:`KeyError`
        '''

        return self._headers[header_name]

    def delete_header(self, header_name):
        '''Deletes the header from the request.

        Parameters
        ----------
        header_name : :obj:`str`
            String name of the header.

        Raises
        ------
        :class:`KeyError`
        '''

        if header_name == 'Compress' and self.body:
            self._decompress_body()
        self._headers.pop(header_name)



#!/usr/bin/env python3

'''Contains the Client class that is used to interact with SPAMD.'''

import asyncio
from functools import wraps
import logging

from aiospamc.exceptions import (BadResponse, ResponseException,
                                 UsageException, DataErrorException, NoInputException, NoUserException,
                                 NoHostException, UnavailableException, InternalSoftwareException, OSErrorException,
                                 OSFileException, CantCreateException, IOErrorException, TemporaryFailureException,
                                 ProtocolException, NoPermissionException, ConfigException, TimeoutException)
from aiospamc.headers import Compress, MessageClass, Remove, Set, User
from aiospamc.parser import parse, ParseError
from aiospamc.requests import Request
from aiospamc.responses import Status


def _add_compress_header(func):
    '''If the class instance's :attribute:`compress` boolean is `True` then the
    :class:`aiospamc.headers.Compress` header is added to the
    :class:`aiospamc.requests.Request` object.'''

    @wraps(func)
    async def wrapper(cls, request):
        if cls.compress and cls.body:
            cls.logger.debug('Added Compress header to request (%s)', id(request))
            request.add_header(Compress())
        return await func(cls, request)

    return wrapper


def _add_user_header(func):
    '''If the class instance's :attribute:`user` boolean is `True` then the
    :class:`aiospamc.headers.User` header is added to the
    :class:`aiospamc.requests.Request` object.'''

    @wraps(func)
    async def wrapper(cls, request):
        if cls.user:
            cls.logger.debug('Added user header for \'%s\' to request (%s)',
                             cls.user,
                             id(request))
            request.add_header(User(cls.user))
        return await func(cls, request)

    return wrapper


class Client:
    '''Client object for interacting with SPAMD.

    Attributes
    ----------
    connection : :class:`aiospamc.connections.ConnectionManager`
        Manager instance to open connections.
    user : :obj:`str`
        Name of the user that SPAMD will run the checks under.
    compress : :obj:`bool`
        If true, the request body will be compressed.
    loop : :class:`asyncio.AbstractEventLoop`
        The asyncio event loop.
    logger : :class:`logging.Logger`
        Logging instance, logs to 'aiospamc.client'
    '''

    def __init__(self,
                 socket_path='/var/run/spamassassin/spamd.sock',
                 host=None,
                 port=783,
                 user=None,
                 compress=False,
                 ssl=False,
                 loop=None):
        '''Client constructor.

        Parameters
        ----------
        socket_path : :obj:`str`, optional
            The path to the Unix socket for the SPAMD service.
        host : :obj:`str`, optional
            Hostname or IP address of the SPAMD service, defaults to localhost.
        port : :obj:`int`, optional
            Port number for the SPAMD service, defaults to 783.
        user : :obj:`str`, optional
            Name of the user that SPAMD will run the checks under.
        compress : :obj:`bool`, optional
            If true, the request body will be compressed.
        ssl : :obj:`bool`, optional
            If true, will enable SSL/TLS for the connection.
        loop : :class:`asyncio.AbstractEventLoop`
            The asyncio event loop.

        Raises
        ------
        ValueError
            Raised if the constructor can't tell if it's using a TCP or a Unix
            domain socket connection.
        '''

        if host and port:
            from aiospamc.connections.tcp_connection import TcpConnectionManager
            self.connection = TcpConnectionManager(host, port)
        elif socket_path:
            from aiospamc.connections.unix_connection import UnixConnectionManager
            self.connection = UnixConnectionManager(socket_path)
        else:
            raise ValueError('Either "host" and "port" or "socket_path" must be specified.')

        self._host = host
        self._port = port
        self._socket_path = socket_path
        self.user = user
        self.compress = compress
        self._ssl = ssl
        self.loop = loop or asyncio.get_event_loop()

        self.parser = parse

        self.logger = logging.getLogger(__name__)
        self.logger.debug('Created instance of %r', self)

    def __repr__(self):
        client_fmt = ('{}(socket_path={}, '
                      'host={}, '
                      'port={}, '
                      'user={}, '
                      'compress={}, '
                      'ssl={})')
        return client_fmt.format(self.__class__.__name__,
                                 repr(self._socket_path),
                                 repr(self._host),
                                 repr(self._port),
                                 repr(self.user),
                                 repr(self.compress),
                                 repr(self._ssl))

    @staticmethod
    def _raise_response_exception(response):
        if response.status_code is Status.EX_OK:
            return
        elif response.status_code is Status.EX_USAGE:
            raise UsageException(response)
        elif response.status_code is Status.EX_DATAERR:
            raise DataErrorException(response)
        elif response.status_code is Status.EX_NOINPUT:
            raise NoInputException(response)
        elif response.status_code is Status.EX_NOUSER:
            raise NoUserException(response)
        elif response.status_code is Status.EX_NOHOST:
            raise NoHostException(response)
        elif response.status_code is Status.EX_UNAVAILABLE:
            raise UnavailableException(response)
        elif response.status_code is Status.EX_SOFTWARE:
            raise InternalSoftwareException(response)
        elif response.status_code is Status.EX_OSERR:
            raise OSErrorException(response)
        elif response.status_code is Status.EX_OSFILE:
            raise OSFileException(response)
        elif response.status_code is Status.EX_CANTCREAT:
            raise CantCreateException(response)
        elif response.status_code is Status.EX_IOERR:
            raise IOErrorException(response)
        elif response.status_code is Status.EX_TEMPFAIL:
            raise TemporaryFailureException(response)
        elif response.status_code is Status.EX_PROTOCOL:
            raise ProtocolException(response)
        elif response.status_code is Status.EX_NOPERM:
            raise NoPermissionException(response)
        elif response.status_code is Status.EX_CONFIG:
            raise ConfigException(response)
        elif response.status_code is Status.EX_TIMEOUT:
            raise TimeoutException(response)
        else:
            raise ResponseException(response)

    @_add_compress_header
    @_add_user_header
    async def send(self, request):
        '''Sends a request to the SPAMD service.

        If the SPAMD service gives a temporary failure response, then

        Parameters
        ----------
        request : :class:`aiospamc.requests.Request`
            Request object to send.

        Returns
        -------
        :class:`aiospamc.responses.Response`

        Raises
        ------
        :class:`aiospamc.exceptions.BadResponse`
            If the response from SPAMD is ill-formed this exception will be
            raised.
        :class:`aiospamc.exceptions.AIOSpamcConnectionFailed`
            Raised if an error occurred when trying to connect.
        :class:`aiospamc.exceptions.UsageException`
            Error in command line usage.
        :class:`aiospamc.exceptions.DataErrorException`
            Error with data format.
        :class:`aiospamc.exceptions.NoInputException`
            Cannot open input.
        :class:`aiospamc.exceptions.NoUserException`
            Addressee unknown.
        :class:`aiospamc.exceptions.NoHostException`
            Hostname unknown.
        :class:`aiospamc.exceptions.UnavailableException`
            Service unavailable.
        :class:`aiospamc.exceptions.InternalSoftwareException`
            Internal software error.
        :class:`aiospamc.exceptions.OSErrorException`
            System error.
        :class:`aiospamc.exceptions.OSFileException`
            Operating system file missing.
        :class:`aiospamc.exceptions.CantCreateException`
            Cannot create output file.
        :class:`aiospamc.exceptions.IOErrorException`
            Input/output error.
        :class:`aiospamc.exceptions.TemporaryFailureException`
            Temporary failure, may reattempt.
        :class:`aiospamc.exceptions.ProtocolException`
            Error in the protocol.
        :class:`aiospamc.exceptions.NoPermissionException`
            Permission denied.
        :class:`aiospamc.exceptions.ConfigException`
            Error in configuration.
        :class:`aiospamc.exceptions.TimeoutException`
            Timeout during connection.
        '''

        self.logger.debug('Sending request (%s)', id(request))
        async with self.connection.new_connection() as connection:
            await connection.send(bytes(request))
            self.logger.debug('Request (%s) successfully sent', id(request))
            data = await connection.receive()

        try:
            try:
                response = self.parser(data)
            except ParseError:
                raise BadResponse
            self._raise_response_exception(response)
        except ResponseException as error:
            self.logger.exception('Exception for request (%s)when composing response: %s',
                                  id(request),
                                  error)
            raise

        self.logger.debug('Received response (%s) for request (%s)',
            id(response),
            id(request))
        return response

    async def check(self, message):
        '''Request the SPAMD service to check a message with a CHECK request.

        The response will contain a 'Spam' header if the message is marked
        as spam as well as the score and threshold.

        SPAMD will perform a scan on the included message.  SPAMD expects an
        RFC 822 or RFC 2822 formatted email.

        Parameters
        ----------
        message : :obj:`str`
            A string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        Returns
        -------
        :class:`aiospamc.responses.Response`
            The response will contain a 'Spam' header if the message is marked
            as spam as well as the score and threshold.

        Raises
        ------
        :class:`aiospamc.exceptions.BadResponse`
            If the response from SPAMD is ill-formed this exception will be
            raised.
        :class:`aiospamc.exceptions.AIOSpamcConnectionFailed`
            Raised if an error occurred when trying to connect.
        :class:`aiospamc.exceptions.UsageException`
            Error in command line usage.
        :class:`aiospamc.exceptions.DataErrorException`
            Error with data format.
        :class:`aiospamc.exceptions.NoInputException`
            Cannot open input.
        :class:`aiospamc.exceptions.NoUserException`
            Addressee unknown.
        :class:`aiospamc.exceptions.NoHostException`
            Hostname unknown.
        :class:`aiospamc.exceptions.UnavailableException`
            Service unavailable.
        :class:`aiospamc.exceptions.InternalSoftwareException`
            Internal software error.
        :class:`aiospamc.exceptions.OSErrorException`
            System error.
        :class:`aiospamc.exceptions.OSFileException`
            Operating system file missing.
        :class:`aiospamc.exceptions.CantCreateException`
            Cannot create output file.
        :class:`aiospamc.exceptions.IOErrorException`
            Input/output error.
        :class:`aiospamc.exceptions.TemporaryFailureException`
            Temporary failure, may reattempt.
        :class:`aiospamc.exceptions.ProtocolException`
            Error in the protocol.
        :class:`aiospamc.exceptions.NoPermissionException`
            Permission denied.
        :class:`aiospamc.exceptions.ConfigException`
            Error in configuration.
        :class:`aiospamc.exceptions.TimeoutException`
            Timeout during connection.
        '''

        request = Request('CHECK', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def headers(self, message):
        '''Request the SPAMD service to check a message with a HEADERS request.

        Parameters
        ----------
        message : :obj:`str`
            A string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        Returns
        -------
        :class:`aiospamc.responses.Response`
            The response will contain a 'Spam' header if the message is marked
            as spam as well as the score and threshold.

            The body will contain the modified headers of the message.

        Raises
        ------
        :class:`aiospamc.exceptions.BadResponse`
            If the response from SPAMD is ill-formed this exception will be
            raised.
        :class:`aiospamc.exceptions.AIOSpamcConnectionFailed`
            Raised if an error occurred when trying to connect.
        :class:`aiospamc.exceptions.UsageException`
            Error in command line usage.
        :class:`aiospamc.exceptions.DataErrorException`
            Error with data format.
        :class:`aiospamc.exceptions.NoInputException`
            Cannot open input.
        :class:`aiospamc.exceptions.NoUserException`
            Addressee unknown.
        :class:`aiospamc.exceptions.NoHostException`
            Hostname unknown.
        :class:`aiospamc.exceptions.UnavailableException`
            Service unavailable.
        :class:`aiospamc.exceptions.InternalSoftwareException`
            Internal software error.
        :class:`aiospamc.exceptions.OSErrorException`
            System error.
        :class:`aiospamc.exceptions.OSFileException`
            Operating system file missing.
        :class:`aiospamc.exceptions.CantCreateException`
            Cannot create output file.
        :class:`aiospamc.exceptions.IOErrorException`
            Input/output error.
        :class:`aiospamc.exceptions.TemporaryFailureException`
            Temporary failure, may reattempt.
        :class:`aiospamc.exceptions.ProtocolException`
            Error in the protocol.
        :class:`aiospamc.exceptions.NoPermissionException`
            Permission denied.
        :class:`aiospamc.exceptions.ConfigException`
            Error in configuration.
        :class:`aiospamc.exceptions.TimeoutException`
            Timeout during connection.
        '''

        request = Request('HEADERS', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def ping(self):
        '''Sends a ping request to the SPAMD service and will receive a
        response if the serivce is alive.

        Returns
        -------
        :class:`aiospamc.responses.Response`
            Response message will contain 'PONG' if successful.

        Raises
        ------
        :class:`aiospamc.exceptions.BadResponse`
            If the response from SPAMD is ill-formed this exception will be
            raised.
        :class:`aiospamc.exceptions.AIOSpamcConnectionFailed`
            Raised if an error occurred when trying to connect.
        :class:`aiospamc.exceptions.UsageException`
            Error in command line usage.
        :class:`aiospamc.exceptions.DataErrorException`
            Error with data format.
        :class:`aiospamc.exceptions.NoInputException`
            Cannot open input.
        :class:`aiospamc.exceptions.NoUserException`
            Addressee unknown.
        :class:`aiospamc.exceptions.NoHostException`
            Hostname unknown.
        :class:`aiospamc.exceptions.UnavailableException`
            Service unavailable.
        :class:`aiospamc.exceptions.InternalSoftwareException`
            Internal software error.
        :class:`aiospamc.exceptions.OSErrorException`
            System error.
        :class:`aiospamc.exceptions.OSFileException`
            Operating system file missing.
        :class:`aiospamc.exceptions.CantCreateException`
            Cannot create output file.
        :class:`aiospamc.exceptions.IOErrorException`
            Input/output error.
        :class:`aiospamc.exceptions.TemporaryFailureException`
            Temporary failure, may reattempt.
        :class:`aiospamc.exceptions.ProtocolException`
            Error in the protocol.
        :class:`aiospamc.exceptions.NoPermissionException`
            Permission denied.
        :class:`aiospamc.exceptions.ConfigException`
            Error in configuration.
        :class:`aiospamc.exceptions.TimeoutException`
            Timeout during connection.
        '''

        request = Request('PING')
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def process(self, message):
        '''Request the SPAMD service to check a message with a PROCESS request.

        Parameters
        ----------
        message : :obj:`str`
            A string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        Returns
        -------
        :class:`aiospamc.responses.Response`
            The response will contain a 'Spam' header if the message is marked
            as spam as well as the score and threshold.

            The body will contain a modified version of the message.

        Raises
        ------
        :class:`aiospamc.exceptions.BadResponse`
            If the response from SPAMD is ill-formed this exception will be
            raised.
        :class:`aiospamc.exceptions.AIOSpamcConnectionFailed`
            Raised if an error occurred when trying to connect.
        :class:`aiospamc.exceptions.UsageException`
            Error in command line usage.
        :class:`aiospamc.exceptions.DataErrorException`
            Error with data format.
        :class:`aiospamc.exceptions.NoInputException`
            Cannot open input.
        :class:`aiospamc.exceptions.NoUserException`
            Addressee unknown.
        :class:`aiospamc.exceptions.NoHostException`
            Hostname unknown.
        :class:`aiospamc.exceptions.UnavailableException`
            Service unavailable.
        :class:`aiospamc.exceptions.InternalSoftwareException`
            Internal software error.
        :class:`aiospamc.exceptions.OSErrorException`
            System error.
        :class:`aiospamc.exceptions.OSFileException`
            Operating system file missing.
        :class:`aiospamc.exceptions.CantCreateException`
            Cannot create output file.
        :class:`aiospamc.exceptions.IOErrorException`
            Input/output error.
        :class:`aiospamc.exceptions.TemporaryFailureException`
            Temporary failure, may reattempt.
        :class:`aiospamc.exceptions.ProtocolException`
            Error in the protocol.
        :class:`aiospamc.exceptions.NoPermissionException`
            Permission denied.
        :class:`aiospamc.exceptions.ConfigException`
            Error in configuration.
        :class:`aiospamc.exceptions.TimeoutException`
            Timeout during connection.
        '''

        request = Request('PROCESS', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def report(self, message):
        '''Request the SPAMD service to check a message with a REPORT request.

        Parameters
        ----------
        message : :obj:`str`
            A string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        Returns
        -------
        :class:`aiospamc.responses.Response`
            The response will contain a 'Spam' header if the message is marked
            as spam as well as the score and threshold.

            The body will contain a report composed by the SPAMD service.

        Raises
        ------
        :class:`aiospamc.exceptions.BadResponse`
            If the response from SPAMD is ill-formed this exception will be
            raised.
        :class:`aiospamc.exceptions.AIOSpamcConnectionFailed`
            Raised if an error occurred when trying to connect.
        :class:`aiospamc.exceptions.UsageException`
            Error in command line usage.
        :class:`aiospamc.exceptions.DataErrorException`
            Error with data format.
        :class:`aiospamc.exceptions.NoInputException`
            Cannot open input.
        :class:`aiospamc.exceptions.NoUserException`
            Addressee unknown.
        :class:`aiospamc.exceptions.NoHostException`
            Hostname unknown.
        :class:`aiospamc.exceptions.UnavailableException`
            Service unavailable.
        :class:`aiospamc.exceptions.InternalSoftwareException`
            Internal software error.
        :class:`aiospamc.exceptions.OSErrorException`
            System error.
        :class:`aiospamc.exceptions.OSFileException`
            Operating system file missing.
        :class:`aiospamc.exceptions.CantCreateException`
            Cannot create output file.
        :class:`aiospamc.exceptions.IOErrorException`
            Input/output error.
        :class:`aiospamc.exceptions.TemporaryFailureException`
            Temporary failure, may reattempt.
        :class:`aiospamc.exceptions.ProtocolException`
            Error in the protocol.
        :class:`aiospamc.exceptions.NoPermissionException`
            Permission denied.
        :class:`aiospamc.exceptions.ConfigException`
            Error in configuration.
        :class:`aiospamc.exceptions.TimeoutException`
            Timeout during connection.
        '''

        request = Request('REPORT', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def report_if_spam(self, message):
        '''Request the SPAMD service to check a message with a REPORT_IFSPAM
        request.

        Parameters
        ----------
        message : :obj:`str`
            A string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        Returns
        -------
        :class:`aiospamc.responses.Response`
            The response will contain a 'Spam' header if the message is marked
            as spam as well as the score and threshold.

            The body will contain a report composed by the SPAMD service only if
            message is marked as being spam.

        Raises
        ------
        :class:`aiospamc.exceptions.BadResponse`
            If the response from SPAMD is ill-formed this exception will be
            raised.
        :class:`aiospamc.exceptions.AIOSpamcConnectionFailed`
            Raised if an error occurred when trying to connect.
        :class:`aiospamc.exceptions.UsageException`
            Error in command line usage.
        :class:`aiospamc.exceptions.DataErrorException`
            Error with data format.
        :class:`aiospamc.exceptions.NoInputException`
            Cannot open input.
        :class:`aiospamc.exceptions.NoUserException`
            Addressee unknown.
        :class:`aiospamc.exceptions.NoHostException`
            Hostname unknown.
        :class:`aiospamc.exceptions.UnavailableException`
            Service unavailable.
        :class:`aiospamc.exceptions.InternalSoftwareException`
            Internal software error.
        :class:`aiospamc.exceptions.OSErrorException`
            System error.
        :class:`aiospamc.exceptions.OSFileException`
            Operating system file missing.
        :class:`aiospamc.exceptions.CantCreateException`
            Cannot create output file.
        :class:`aiospamc.exceptions.IOErrorException`
            Input/output error.
        :class:`aiospamc.exceptions.TemporaryFailureException`
            Temporary failure, may reattempt.
        :class:`aiospamc.exceptions.ProtocolException`
            Error in the protocol.
        :class:`aiospamc.exceptions.NoPermissionException`
            Permission denied.
        :class:`aiospamc.exceptions.ConfigException`
            Error in configuration.
        :class:`aiospamc.exceptions.TimeoutException`
            Timeout during connection.
        '''

        request = Request('REPORT_IFSPAM', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def symbols(self, message):
        '''Request the SPAMD service to check a message with a SYMBOLS request.

        The response will contain a 'Spam' header if the message is marked
        as spam as well as the score and threshold.

        Parameters
        ----------
        message : :obj:`str`
            A string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.

        Returns
        -------
        :class:`aiospamc.responses.Response`
            Will contain a 'Spam' header if the message is marked as spam as
            well as the score and threshold.

            The body will contain a comma separated list of all the rule names.

        Raises
        ------
        :class:`aiospamc.exceptions.BadResponse`
            If the response from SPAMD is ill-formed this exception will be
            raised.
        :class:`aiospamc.exceptions.AIOSpamcConnectionFailed`
            Raised if an error occurred when trying to connect.
        :class:`aiospamc.exceptions.UsageException`
            Error in command line usage.
        :class:`aiospamc.exceptions.DataErrorException`
            Error with data format.
        :class:`aiospamc.exceptions.NoInputException`
            Cannot open input.
        :class:`aiospamc.exceptions.NoUserException`
            Addressee unknown.
        :class:`aiospamc.exceptions.NoHostException`
            Hostname unknown.
        :class:`aiospamc.exceptions.UnavailableException`
            Service unavailable.
        :class:`aiospamc.exceptions.InternalSoftwareException`
            Internal software error.
        :class:`aiospamc.exceptions.OSErrorException`
            System error.
        :class:`aiospamc.exceptions.OSFileException`
            Operating system file missing.
        :class:`aiospamc.exceptions.CantCreateException`
            Cannot create output file.
        :class:`aiospamc.exceptions.IOErrorException`
            Input/output error.
        :class:`aiospamc.exceptions.TemporaryFailureException`
            Temporary failure, may reattempt.
        :class:`aiospamc.exceptions.ProtocolException`
            Error in the protocol.
        :class:`aiospamc.exceptions.NoPermissionException`
            Permission denied.
        :class:`aiospamc.exceptions.ConfigException`
            Error in configuration.
        :class:`aiospamc.exceptions.TimeoutException`
            Timeout during connection.
        '''

        request = Request('SYMBOLS', body=message)
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response

    async def tell(self,
                   message_class,
                   message,
                   remove_action=None,
                   set_action=None,
                  ):
        '''Instruct the SPAMD service to to mark the message

        Parameters
        ----------
        message_class : :class:`aiospamc.options.MessageClassOption`
            An enumeration to classify the message as 'spam' or 'ham.'
        message : :obj:`str`
            A string containing the contents of the message to be scanned.

            SPAMD will perform a scan on the included message.  SPAMD expects an
            RFC 822 or RFC 2822 formatted email.
        remove_action : :class:`aiospamc.options.ActionOption`
            Remove message class for message in database.
        set_action : :class:`aiospamc.options.ActionOption`
            Set message class for message in database.

        Returns
        -------
        :class:`aiospamc.responses.Response`
            Will contain a 'Spam' header if the message is marked as spam as
            well as the score and threshold.

            The body will contain a report composed by the SPAMD service only if
            message is marked as being spam.

        Raises
        ------
        :class:`aiospamc.exceptions.BadResponse`
            If the response from SPAMD is ill-formed this exception will be
            raised.
        :class:`aiospamc.exceptions.AIOSpamcConnectionFailed`
            Raised if an error occurred when trying to connect.
        :class:`aiospamc.exceptions.UsageException`
            Error in command line usage.
        :class:`aiospamc.exceptions.DataErrorException`
            Error with data format.
        :class:`aiospamc.exceptions.NoInputException`
            Cannot open input.
        :class:`aiospamc.exceptions.NoUserException`
            Addressee unknown.
        :class:`aiospamc.exceptions.NoHostException`
            Hostname unknown.
        :class:`aiospamc.exceptions.UnavailableException`
            Service unavailable.
        :class:`aiospamc.exceptions.InternalSoftwareException`
            Internal software error.
        :class:`aiospamc.exceptions.OSErrorException`
            System error.
        :class:`aiospamc.exceptions.OSFileException`
            Operating system file missing.
        :class:`aiospamc.exceptions.CantCreateException`
            Cannot create output file.
        :class:`aiospamc.exceptions.IOErrorException`
            Input/output error.
        :class:`aiospamc.exceptions.TemporaryFailureException`
            Temporary failure, may reattempt.
        :class:`aiospamc.exceptions.ProtocolException`
            Error in the protocol.
        :class:`aiospamc.exceptions.NoPermissionException`
            Permission denied.
        :class:`aiospamc.exceptions.ConfigException`
            Error in configuration.
        :class:`aiospamc.exceptions.TimeoutException`
            Timeout during connection.
        '''

        request = Request(verb='TELL',
                          headers=(MessageClass(message_class),),
                          body=message)
        if remove_action:
            request.add_header(Set(remove_action))
        if set_action:
            request.add_header(Remove(set_action))
        self.logger.debug('Composed %s request (%s)', request.verb, id(request))
        response = await self.send(request)

        return response



#!/usr/bin/env python3

'''Contains classes used for responses.'''

import enum

from aiospamc.common import RequestResponseBase
from aiospamc.exceptions import (UsageException, DataErrorException, NoInputException, NoUserException,
                                 NoHostException, UnavailableException, InternalSoftwareException, OSErrorException,
                                 OSFileException, CantCreateException, IOErrorException, TemporaryFailureException,
                                 ProtocolException, NoPermissionException, ConfigException, TimeoutException)


class Status(enum.IntEnum):
    '''Enumeration of status codes that the SPAMD will accompany with a
    response.

    Reference: https://svn.apache.org/repos/asf/spamassassin/trunk/spamd/spamd.raw
    Look for the %resphash variable.
    '''

    #pylint: disable=C0326
    def __new__(cls, value, exception=None, description=''):
        #pylint: disable=protected-access
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj.exception = exception
        obj.description = description

        return obj

    EX_OK           = 0,    None,          'No problems'
    EX_USAGE        = 64, UsageException, 'Command line usage error'
    EX_DATAERR      = 65, DataErrorException, 'Data format error'
    EX_NOINPUT      = 66, NoInputException, 'Cannot open input'
    EX_NOUSER       = 67, NoUserException, 'Addressee unknown'
    EX_NOHOST       = 68, NoHostException, 'Host name unknown'
    EX_UNAVAILABLE  = 69, UnavailableException, 'Service unavailable'
    EX_SOFTWARE     = 70, InternalSoftwareException, 'Internal software error'
    EX_OSERR        = 71, OSErrorException, 'System error (e.g., can\'t fork)'
    EX_OSFILE       = 72, OSFileException, 'Critical OS file missing'
    EX_CANTCREAT    = 73, CantCreateException, 'Can\'t create (user) output file'
    EX_IOERR        = 74, IOErrorException, 'Input/output error'
    EX_TEMPFAIL     = 75, TemporaryFailureException, 'Temp failure; user is invited to retry'
    EX_PROTOCOL     = 76, ProtocolException, 'Remote error in protocol'
    EX_NOPERM       = 77, NoPermissionException, 'Permission denied'
    EX_CONFIG       = 78, ConfigException, 'Configuration error'
    EX_TIMEOUT      = 79, TimeoutException, 'Read timeout'


class Response(RequestResponseBase):
    '''Class to encapsulate response.

    Attributes
    ----------
    protocol_version : :obj:`str`
        Protocol version given by the response.
    status_code : :class:`aiospamc.responess.Status`
        Status code give by the response.
    message : :obj:`str`
        Message accompanying the status code.
    body : :obj:`str` or :obj:`bytes`
        Contents of the response body.
    '''

    def __init__(self, version, status_code, message, headers=None, body=None):
        '''Response constructor.

        Parameters
        ----------
        version : :obj:`str`
            Version reported by the SPAMD service response.
        status_code : :class:`aiospamc.responses.Status`
            Success or error code.
        message : :obj:`str`
            Message associated with status code.
        body : :obj:`str` or :obj:`bytes`, optional
            String representation of the body.  An instance of the
            :class:`aiospamc.headers.ContentLength` will be automatically added.
        headers : tuple of :class:`aiospamc.headers.Header`, optional
            Collection of headers to be added.  If it contains an instance of
            :class:`aiospamc.headers.Compress` then the body is automatically
            compressed.
        '''

        self.version = version
        self.status_code = status_code
        self.message = message
        super().__init__(body, headers)

    def __bytes__(self):
        if self._compressed_body:
            body = self._compressed_body
        elif self.body:
            body = self.body.encode()
        else:
            body = b''

        return (b'SPAMD/%(version)b '
                b'%(status)d '
                b'%(message)b\r\n'
                b'%(headers)b\r\n'
                b'%(body)b') % {b'version': b'1.5',
                                b'status': self.status_code.value,
                                b'message': self.message.encode(),
                                b'headers': b''.join(map(bytes, self._headers.values())),
                                b'body': body}



#!/usr/bin/env python

'''Connection and ConnectionManager base classes.'''

import asyncio
import logging


class Connection:
    '''Base class for connection objects.

    Attributes
    ----------
    connected : bool
        Status on if the connection is established.
    loop : asyncio.AbstratEventLoop
        The asyncio event loop.
    logger : logging.Logger
        Logging instance.  Logs to 'aiospamc.connections'
    '''

    def __init__(self, loop=None):
        '''Connection constructor.

        Parameters
        ----------
        loop : asyncio.AbstractEventLoop
            The asyncio event loop.
        '''

        self.connected = False
        self.loop = loop or asyncio.get_event_loop()
        self.logger = logging.getLogger(__name__)

    async def __aenter__(self):
        self.logger.debug('Connecting to %s', self.connection_string)
        self.reader, self.writer = await self.open()
        self.connected = True
        self.logger.debug('Connected to %s', self.connection_string)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.logger.debug('Closing connection to %s', self.connection_string)
        self.close()
        self.logger.debug('Closed connection to %s', self.connection_string)

    async def open(self):
        '''Connect to a service.

        Returns
        -------
        asyncio.StreamReader
            Instance of stream reader.
        asyncio.StreamWriter
            Instance of stream writer.

        Raises
        ------
        aiospamc.exceptions.AIOSpamcConnectionFailed
            If connection failed for some reason.
        '''

        raise NotImplementedError

    @property
    def connection_string(self):
        '''String representation of the connection.

        Returns
        -------
        str
            String of the connection address.
        '''

        raise NotImplementedError

    def close(self):
        '''Closes the connection.'''

        self.writer.close()
        self.connected = False
        del self.reader, self.writer

    async def send(self, data):
        '''Sends data through the connection.

        Parameters
        ----------
        data : bytes
            Data to send.
        '''

        self.writer.write(data)
        await self.writer.drain()

    async def receive(self):
        '''Receives data from the connection.

        Returns
        -------
        bytes
            Data received.
        '''

        return await self.reader.read()


class ConnectionManager:
    '''Stores connection parameters and creates connections.
    
    Attributes
    ----------
    loop : asyncio.AbstratEventLoop
        The asyncio event loop.
    '''

    def __init__(self, loop=None):
        self.loop = loop or asyncio.get_event_loop()

    def new_connection(self):
        '''Creates a connection object.

        Returns
        -------
        Connection
            Instance of a Connection object.
        '''

        raise NotImplementedError



#!/usr/bin/env python3

'''Collection of request and response headers.'''

import getpass

from aiospamc.options import ActionOption, MessageClassOption


class Header:
    '''Header base class.'''

    def __bytes__(self):
        raise NotImplementedError

    def __len__(self):
        return len(bytes(self))

    def field_name(self):
        '''Returns the the field name for the header.

        Returns
        -------
        str
        '''

        raise NotImplementedError


class Compress(Header):
    '''Compress header.  Specifies what encryption scheme to use.  So far only
    'zlib' is supported.

    Attributes
    ----------
    zlib : bool
        True if the zlib compression algorithm is used.
    '''

    def __init__(self):
        self.zlib = True

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)

    def __bytes__(self):
        return b'%b: zlib\r\n' % (self.field_name().encode())

    def field_name(self):
        return 'Compress'


class ContentLength(Header):
    '''ContentLength header.  Indicates the length of the body in bytes.

    Attributes
    ----------
    length : int
        Length of the body.
    '''

    def __init__(self, length=0):
        '''ContentLength constructor.

        Parameters
        ----------
        length : :obj:`int`, optional
            Length of the body.
        '''
        self.length = length

    def __bytes__(self):
        return b'%b: %d\r\n' % (self.field_name().encode(),
                                self.length)

    def __repr__(self):
        return '{}(length={})'.format(self.__class__.__name__, self.length)

    def field_name(self):
        return 'Content-length'

class MessageClass(Header):
    '''MessageClass header.  Used to specify whether a message is 'spam' or
    'ham.'

    Attributes
    ----------
    value : :class:`aiospamc.options.MessageClassOption`
        Specifies the classification of the message.
    '''

    def __init__(self, value=None):
        '''MessageClass constructor.

        Parameters
        ----------
        value : :class:`aiospamc.options.MessageClassOption`, optional
            Specifies the classification of the message.
        '''

        self.value = value or MessageClassOption.ham

    def __bytes__(self):
        return b'%b: %b\r\n' % (self.field_name().encode(),
                                self.value.name.encode())

    def __repr__(self):
        return '{}(value={})'.format(self.__class__.__name__, str(self.value))

    def field_name(self):
        return 'Message-class'


class _SetRemoveBase(Header):
    '''Base class for headers that implement "local" and "remote" rules.

    Attributes
    ----------
    action : :class:`aiospamc.options.ActionOption`
        Actions to be done on local or remote.
    '''

    def __init__(self, action=None):
        '''_SetRemoveBase constructor.

        Parameters
        ----------
        action : :class:`aiospamc.options.ActionOption`, optional
            Actions to be done on local or remote.
        '''

        self.action = action or ActionOption(local=False, remote=False)

    def __bytes__(self):
        if not self.action.local and not self.action.remote:
            # if nothing is set, then return a blank string so the request
            # doesn't get tainted
            return b''

        values = []
        if self.action.local:
            values.append(b'local')
        if self.action.remote:
            values.append(b'remote')

        return b'%b: %b\r\n' % (self.field_name().encode(),
                                b', '.join(values))

    def __repr__(self):
        return '{}(action={})'.format(self.__class__.__name__,
                                      repr(self.action))


class DidRemove(_SetRemoveBase):
    '''DidRemove header.  Used by SPAMD to indicate if a message was removed
    from either a local or remote database in response to a TELL request.

    Attributes
    ----------
    action : :class:`aiospamc.options.ActionOption`
        Actions to be done on local or remote.
    '''

    def field_name(self):
        return 'DidRemove'


class DidSet(_SetRemoveBase):
    '''DidRemove header.  Used by SPAMD to indicate if a message was added to
    either a local or remote database in response to a TELL request.

    Attributes
    ----------
    action : :class:`aiospamc.options.ActionOption`
        Actions to be done on local or remote.
    '''

    def field_name(self):
        return 'DidSet'


class Remove(_SetRemoveBase):
    '''Remove header.  Used in a TELL request to ask the SPAMD service remove
    a message from a local or remote database.  The SPAMD service must have the
    --allow-tells switch in order for this to do anything.

    Attributes
    ----------
    action : :class:`aiospamc.options.ActionOption`
        Actions to be done on local or remote.
    '''

    def field_name(self):
        return 'Remove'


class Set(_SetRemoveBase):
    '''Set header.  Used in a TELL request to ask the SPAMD service add a
    message from a local or remote database.  The SPAMD service must have the
    --allow-tells switch in order for this to do anything.

    Attributes
    ----------
    action : :class:`aiospamc.options.ActionOption`
        Actions to be done on local or remote.
    '''

    def field_name(self):
        return 'Set'


class Spam(Header):
    '''Spam header.  Used by the SPAMD service to report on if the submitted
    message was spam and the score/threshold that it used.

    Attributes
    ----------
    value : :obj:`bool`
            True if the message is spam, False if not.
    score : :obj:`float`
        Score of the message after being scanned.
    threshold : :obj:`float`
        Threshold of which the message would have been marked as spam.
    '''

    def __init__(self, value=False, score=0.0, threshold=0.0):
        '''Spam header constructor.

        Parameters
        ----------
        value : :obj:`bool`, optional
            True if the message is spam, False if not.
        score : :obj:`float`, optional
            Score of the message after being scanned.
        threshold : :obj:`float`, optional
            Threshold of which the message would have been marked as spam.
        '''

        self.value = value
        self.score = score
        self.threshold = threshold

    def __bytes__(self):
        return b'%b: %b ; %.1f / %.1f\r\n' % (self.field_name().encode(),
                                              str(self.value).encode(),
                                              self.score,
                                              self.threshold)

    def __repr__(self):
        return '{}(value={}, score={}, threshold={})'.format(self.__class__.__name__,
                                                             self.value,
                                                             self.score,
                                                             self.threshold)

    def field_name(self):
        return 'Spam'


class User(Header):
    '''User header.  Used to specify which user the SPAMD service should use
    when loading configuration files.

    Attributes
    ----------
    name : :obj:`str`
        Name of the user account.
    '''

    def __init__(self, name=None):
        '''User constructor.

        Parameters
        ----------
        name : :obj:`str`, optional
            Name of the user account.
        '''

        self.name = name or getpass.getuser()

    def __bytes__(self):
        return b'%b: %b\r\n' % (self.field_name().encode(),
                                self.name.encode())

    def __repr__(self):
        return '{}(name={})'.format(self.__class__.__name__, repr(self.name))

    def field_name(self):
        return 'User'


class XHeader(Header):
    '''Extension header.  Used to specify a header that's not supported
    natively by the SPAMD service.

    Attributes
    ----------
    name : :obj:`str`
        Name of the header.
    value : :obj:`str`
        Contents of the value.
    '''

    def __init__(self, name, value):
        '''XHeader constructor.

        Parameters
        ----------
        name : :obj:`str`
            Name of the header.
        value : :obj:`str`
            Contents of the value.
        '''

        self.name = name
        self.value = value

    def __bytes__(self):
        return b'%b: %b\r\n' % (self.field_name().encode(),
                                self.value.encode())

    def __repr__(self):
        return '{}(name={}, value={})'.format(self.__class__.__name__,
                                              repr(self.name),
                                              repr(self.value))

    def field_name(self):
        return self.name



#!/usr/bin/env python3

'''Contains all requests that can be made to the SPAMD service.'''

from aiospamc.common import RequestResponseBase


class Request(RequestResponseBase):
    '''SPAMC request object.

    Attributes
    ----------
    verb : :obj:`str`
        Method name of the request.
    version : :obj:`str`
        Protocol version.
    body : :obj:`str` or :obj:`bytes`
        String representation of the body.  An instance of the
        :class:`aiospamc.headers.ContentLength` will be automatically added.
    '''

    def __init__(self, verb, version='1.5', headers=None, body=None):
        '''Request constructor.

        Parameters
        ----------
        verb : :obj:`str`
            Method name of the request.
        version: :obj:`str`
            Version of the protocol.
        body : :obj:`str` or :obj:`bytes`, optional
            String representation of the body.  An instance of the
            :class:`aiospamc.headers.ContentLength` will be automatically added.
        headers : tuple of :class:`aiospamc.headers.Header`, optional
            Collection of headers to be added.  If it contains an instance of
            :class:`aiospamc.headers.Compress` then the body is automatically
            compressed.
        '''

        self.verb = verb
        self.version = version
        super().__init__(body, headers)

    def __bytes__(self):
        if self._compressed_body:
            body = self._compressed_body
        elif self.body:
            body = self.body.encode()
        else:
            body = b''

        request = (b'%(verb)b '
                   b'SPAMC/%(version)b'
                   b'\r\n'
                   b'%(headers)b\r\n'
                   b'%(body)b')

        return request % {b'verb': self.verb.encode(),
                          b'version': self.version.encode(),
                          b'headers': b''.join(map(bytes, self._headers.values())),
                          b'body': body}



#!/usr/bin/env python3

'''Parser object for SPAMC/SPAMD requests and responses.'''


import re
from functools import wraps

import aiospamc.headers
import aiospamc.options
import aiospamc.requests
import aiospamc.responses


class ParseError(Exception):
    '''An exception occurring when parsing.

    Attributes
    ----------
    index : :obj:`int`
        Index in the stream the exception occurred.
    message : :obj:`str`
        User readable message.
    '''

    def __init__(self, index, message):
        '''ParseError constructor.

        Parameters
        ----------
        index : :obj:`int`
            Index in the stream the exception occurred.
        message : :obj:`str`
            User readable message.
        '''
        self.index = index
        self.message = message


def checkpoint(func):
    '''A decorator to restore the index if an exception occurred.'''

    @wraps(func)
    def inner(self):
        check = self.index
        try:
            return func(self)
        except ParseError:
            self.index = check
            raise
    return inner


class Parser:
    '''Parser object for requests and responses.'''

    def __init__(self, string=None):
        '''Parser constructor.

        Parameters
        ----------
        string : :obj:`str`, optional
            The string to parse.
        '''
        self.string = string or b''
        self.index = 0

    def advance(self, by):
        '''Advance the current index by number of bytes.

        Parameters
        ----------
        by : :obj:`int`
            Number of bytes in the stream to advance.
        '''

        self.index += by

    def current(self):
        '''The remainder of the string that hasn't been parsed.

        Returns
        -------
        :obj:`str`
        '''

        return self.string[self.index:]

    def end(self):
        '''Whether the parser has parsed the entire string.

        Returns
        -------
        :obj:`bool`
        '''

        return self.index >= len(self.string)

    def match(self, pattern):
        '''Returns the regular expression matches string at the current index.

        Parameters
        ----------
        pattern : :obj:`bytes`

        Returns
        -------
        Regular expression match
        '''

        success = re.match(pattern, self.current())
        if success:
            return success
        else:
            return None

    def consume(self, pattern):
        '''If the pattern matches, advances the index the length of the match.
        Returns the regular expression match.

        Parameters
        ----------
        pattern : :obj:`bytes`

        Returns
        -------
        Regular expression match

        Raises
        ------
        :class:`aiospamc.parser.ParseError`
        '''

        match = self.match(pattern)
        if match:
            self.advance(match.end())
            return match
        else:
            raise ParseError(index=self.index, message='Unable to consume match  with pattern {}'.format(pattern))

    @staticmethod
    def skip(func):
        '''Makes the parser function optional by ignore whether it raises a
        :class:`aiospamc.parser.ParseError` exception or not.

        Parameters
        ----------
        func : function
            Function to execute.
        '''

        try:
            func()
        except ParseError:
            pass

    def whitespace(self):
        '''Consumes spaces or tabs.'''

        self.consume(rb'[ \t]+')

    def newline(self):
        '''Consumes a newline sequence (carriage return and line feed).'''

        self.consume(rb'\r\n')

    def version(self):
        '''Consumes a version pattern.  For example, "1.5".

        Returns
        -------
        :obj:`str`
        '''

        return self.consume(rb'\d+\.\d+').group().decode()

    def body(self):
        '''Consumes the rest of the message and returns the contents.

        Returns
        -------
        :obj:`bytes`
        '''

        return self.current()

    # Header functions

    @checkpoint
    def compress_value(self):
        '''Consumes the Compression header value.

        Returns
        -------
        :obj:`str`
        '''

        self.skip(self.whitespace)
        algorithm = self.consume(rb'zlib').group()
        self.skip(self.whitespace)

        return algorithm.decode()

    @checkpoint
    def content_length_value(self):
        '''Consumes the Content-length header value.

        Returns
        -------
        :obj:`int`
        '''

        self.skip(self.whitespace)
        length = int(self.consume(rb'\d+').group())
        self.skip(self.whitespace)

        return length

    @checkpoint
    def message_class_value(self):
        '''Consumes the Message-class header value.

        Returns
        -------
        :class:`aiospamc.options.MessageClassOption`
        '''

        self.skip(self.whitespace)
        m_class = aiospamc.options.MessageClassOption.ham if self.consume(
            rb'(ham|spam)').group() == b'ham' else aiospamc.options.MessageClassOption.spam
        self.skip(self.whitespace)

        return m_class

    @checkpoint
    def set_remove_value(self):
        '''Consumes the value for the DidRemove, DidSet, Remove and Set headers.

        Returns
        -------
        :class:`aiospamc.options.ActionOption`
        '''

        self.skip(self.whitespace)
        action = self.consume(rb'(local|remote)([ \t]*,[ \t]*(local|remote))?').group()
        self.skip(self.whitespace)

        return aiospamc.options.ActionOption(local=b'local' in action, remote=b'remote' in action)

    @checkpoint
    def spam_value(self):
        '''Consumes the Spam header value.

        Returns
        -------
        :obj:`dict`
            Has the keys `value`, `score`, and `threshold`.
        '''

        number = rb'(\d+(\.\d+)?)'

        self.skip(self.whitespace)
        value = True if self.consume(rb'(True|False)').group() == b'True' else False
        self.skip(self.whitespace)
        self.consume(rb';')
        self.skip(self.whitespace)
        score = float(self.consume(number).group())
        self.skip(self.whitespace)
        self.consume(rb'/')
        self.skip(self.whitespace)
        threshold = float(self.consume(number).group())
        self.skip(self.whitespace)

        return {'value': value, 'score': score, 'threshold': threshold}

    @checkpoint
    def user_value(self):
        '''Consumes the User header value.

        Returns
        -------
        :obj:`str`
        '''

        self.skip(self.whitespace)
        username = self.consume(rb'[a-zA-Z0-9-_]+').group()
        self.skip(self.whitespace)

        return username.decode()

    @checkpoint
    def header(self):
        '''Consumes the string and returns an instance of
        :class:`aiospamc.headers.Header`.

        Returns
        -------
        :class:`aiospamc.headers.Header`
        '''

        self.skip(self.whitespace)
        name = self.consume(rb'[a-zA-Z0-9_-]+').group()
        self.skip(self.whitespace)
        self.consume(rb':')
        if name == b'Compress':
            self.compress_value()
            return aiospamc.headers.Compress()
        elif name == b'Content-length':
            return aiospamc.headers.ContentLength(length=self.content_length_value())
        elif name == b'DidRemove':
            return aiospamc.headers.DidRemove(action=self.set_remove_value())
        elif name == b'DidSet':
            return aiospamc.headers.DidSet(action=self.set_remove_value())
        elif name == b'Message-class':
            return aiospamc.headers.MessageClass(value=self.message_class_value())
        elif name == b'Remove':
            return aiospamc.headers.Remove(action=self.set_remove_value())
        elif name == b'Set':
            return aiospamc.headers.Set(action=self.set_remove_value())
        elif name == b'Spam':
            return aiospamc.headers.Spam(**self.spam_value())
        elif name == b'User':
            return aiospamc.headers.User(name=self.user_value())
        else:
            return aiospamc.headers.XHeader(
                    name=name.decode(),
                    value=self.consume(rb'.+(?=\r\n)').group().decode()
            )

    def headers(self):
        '''Consumes all headers.

        Returns
        -------
        :obj:`list` of :class:`aiospamc.headers.Header`
        '''

        header_list = []
        while not self.match(rb'\r\n'):
            try:
                h = self.header()
                self.newline()
                header_list.append(h)
            except ParseError:
                break

        return header_list

    # Request functions

    def spamc_protocol(self):
        '''Consumes the string "SPAMC".

        Returns
        -------
        :obj:`str`
        '''

        return self.consume(rb'SPAMC').group().decode()

    @checkpoint
    def method(self):
        '''Consumes the method name in a request.

        Returns
        -------
        :obj:`str`
        '''

        name = self.consume(rb'CHECK|'
                            rb'HEADERS|'
                            rb'PING|'
                            rb'PROCESS|'
                            rb'REPORT_IFSPAM|'
                            rb'REPORT|'
                            rb'SKIP|'
                            rb'SYMBOLS|'
                            rb'TELL')
        self.whitespace()

        return name.group().decode()

    @checkpoint
    def request(self):
        '''Consumes a SPAMC request.

        Returns
        -------
        :class:`aiospamc.requests.Request`
        '''

        m = self.method()
        p = self.spamc_protocol()
        self.consume(rb'/')
        v = self.version()
        self.newline()

        if not self.end():
            h = self.headers()
        else:
            h = []

        if not self.end():
            self.newline()
            b = self.body()
        else:
            b = None

        return aiospamc.requests.Request(verb=m, version=v, headers=h, body=b)

    # Response functions

    def spamd_protocol(self):
        '''Consumes the string "SPAMD".

        Returns
        -------
        :obj:`str`
        '''

        return self.consume(rb'SPAMD').group().decode()

    @checkpoint
    def status_code(self):
        '''Consumes the status code.

        Returns
        -------
        :class:`aiospamc.responses.StatusCode` or int
        '''

        code = int(self.consume(rb'\d+').group())
        try:
            return aiospamc.responses.Status(code)
        except ValueError:
            return code

    def message(self):
        '''Consumes a string until it matches a newline.

        Returns
        -------
        :obj:`str`
        '''

        return self.consume(rb'.*(?=\r\n)').group().decode()

    @checkpoint
    def response(self):
        '''Consumes a SPAMD response.

        Returns
        -------
        :class:`aiospamc.responses.Response`
        '''

        p = self.spamd_protocol()
        self.consume(rb'/')
        v = self.version()
        self.whitespace()
        c = self.status_code()
        self.whitespace()
        m = self.message()
        self.newline()

        if not self.end():
            h = self.headers()
        else:
            h = []

        if not self.end():
            self.newline()
            b = self.body()
        else:
            b = None

        return aiospamc.responses.Response(version=v, status_code=c, message=m, headers=h,
                                           body=b)

def parse(string):
    '''Parses a request or response.

    Returns
    -------
    :class:`aiospamc.requests.Request` or :class:`aiospamc.responses.Response`
    '''

    parser = Parser(string)

    try:
        return parser.request()
    except ParseError:
        pass

    try:
        return parser.response()
    except ParseError:
        pass

    raise ParseError(0, 'Unable to parse request or response')



#!/usr/bin/env python3

'''aiospamc package.

An asyncio-based library to communicate with SpamAssassin's SPAMD service.'''

from aiospamc.client import Client
from aiospamc.options import ActionOption, MessageClassOption

__all__ = ('Client',
           'MessageClassOption',
           'ActionOption')

__author__ = 'Michael Caley'
__copyright__ = 'Copyright 2016-2018 Michael Caley'
__license__ = 'MIT'
__version__ = '0.4.1'
__email__ = 'mjcaley@darkarctic.com'



#!/usr/bin/env python3

'''setuptools module for aiospamc.'''

from pathlib import Path
from setuptools import setup, find_packages


readme = Path(__file__).absolute().parent / 'README.rst'
with readme.open(encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='aiospamc',
    version='0.4.1',

    description='An asyncio-based library to communicate with SpamAssassin\'s '
                'SPAMD service.',
    long_description=long_description,

    url='https://github.com/mjcaley/aiospamc',

    author='Michael Caley',
    author_email='mjcaley@darkarctic.com',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',

        'License :: OSI Approved :: MIT License',

        'Topic :: Communications :: Email :: Filters',
    ],

    keywords='spam spamc spamassassin',

    packages=find_packages(exclude=['tests']),

    python_requires='!=2.*,!=3.0,!=3.1,!=3.2,!=3.3,!=3.4',
    setup_requires=['pytest-runner', ],
    tests_require=['pytest-asyncio>=0.6', 'pytest-cov', 'pytest>=3.0', 'asynctest'],
)



#!/usr/bin/env python3

'''TCP socket connection and manager.'''

import asyncio

from aiospamc.connections import Connection
from aiospamc.connections import ConnectionManager
from aiospamc.exceptions import AIOSpamcConnectionFailed


class TcpConnectionManager(ConnectionManager):
    '''Creates new connections based on host and port provided.

    Attributes
    ----------
    host : str
        Hostname or IP address of server.
    port : str
        Port number.
    ssl : bool
        Whether to use SSL/TLS.
    '''

    def __init__(self, host, port, ssl=False, loop=None):
        '''Constructor for TcpConnectionManager.

        Parameters
        ----------
        host : str
            Hostname or IP address of server.
        port : str
            Port number
        ssl : :obj:`bool` or optional
            SSL/TLS enabled.
        loop : asyncio.AbstractEventLoop
            The asyncio event loop.
        '''

        self.host = host
        self.port = port
        self.ssl = ssl
        super().__init__(loop)

    def __repr__(self):
        return '{}(host={}, port={}, ssl={})'.format(self.__class__.__name__,
                                                     repr(self.host),
                                                     repr(self.port),
                                                     self.ssl)

    def new_connection(self):
        '''Creates a new TCP connection.

        Raises
        ------
        aiospamc.exceptions.AIOSpamcConnectionFailed
        '''

        return TcpConnection(self.host, self.port, self.ssl, self.loop)


class TcpConnection(Connection):
    '''Manages a TCP connection.

    Attributes
    ----------
    host : str
        Hostname or IP address of server.
    port : str
        Port number
    ssl : bool
        Whether to use SSL/TLS.
    loop : asyncio.AbstratEventLoop
        The asyncio event loop.
    '''

    def __init__(self, host, port, ssl, loop=None):
        '''Constructor for TcpConnection.

        Attributes
        ----------
        host : str
            Hostname or IP address of server.
        port : str
            Port number
        ssl :  :obj:`bool` or optional
            SSL/TLS enabled.
        '''

        self.host = host
        self.port = port
        self.ssl = ssl
        super().__init__(loop)

    def __repr__(self):
        return '{}(host={}, port={}, ssl={})'.format(self.__class__.__name__,
                                                     repr(self.host),
                                                     repr(self.port),
                                                     self.ssl)

    async def open(self):
        '''Opens a connection.

        Returns
        -------
        asyncio.StreamReader
        asyncio.StreamWriter

        Raises
        ------
        aiospamc.exceptions.AIOSpamcConnectionFailed
        '''

        try:
            reader, writer = await asyncio.open_connection(self.host,
                                                           self.port,
                                                           ssl=self.ssl,
                                                           loop=self.loop)
        except (ConnectionRefusedError, OSError) as error:
            raised = AIOSpamcConnectionFailed(error)
            self.logger.exception('Exception occurred when connecting to %s:%s: %s',
                                  self.host,
                                  self.port,
                                  raised)
            raise raised

        return reader, writer

    @property
    def connection_string(self):
        '''String representation of the connection.

        Returns
        -------
        str
            Hostname and port.
        '''

        return ':'.join([self.host, str(self.port)])
